#include "math3d.h"

/////////////////////////////////////////////////
//
//  Implementations of vector member functions,
//  these should be used in code that need speed over
//  size & readability
//
/////////////////////////////////////////////////

void Vector3d::operator+=(Vector3d &vec)
{
    x+=vec.x;
    y+=vec.y;
    z+=vec.z;
}


void Vector3d::operator-=(Vector3d &vec)
{
    x-=vec.x;
    y-=vec.y;
    z-=vec.z;     
}


void Vector3d::operator*=(float scalar)
{
    x*=scalar;
    y*=scalar;
    z*=scalar;   
}


void Vector3d::operator/=(float scalar)
{
    x/=scalar;
    y/=scalar;
    z/=scalar;      
}


float Vector3d::Length()
{
    return sqrtf(x*x+y*y+z*z);
}


void Vector3d::Normalize()
{
    float mult=1.0f/sqrtf(x*x+y*y+z*z);
    x*=mult;
    y*=mult;
    z*=mult;
}


void Vector3d::Invert()
{
    x=-x;
    y=-y;
    z=-y;
}


float Vector3d::Dot(Vector3d &vec)
{
    return x*vec.x+y*vec.y+z*vec.z;
}


Vector3d Vector3d::Cross(Vector3d &vec)
{
    return Vector3d(y*vec.z-vec.y*z,
                    z*vec.x-vec.z*x,
                    x*vec.y-vec.x*y);
}


void Vector3d::MultiplyBy(Matrix3x3 &mat)
{
    float *M=mat.M;
    float X,Y,Z;        
                       
    X=M[0]*x+M[3]*y+M[6]*z; 
    Y=M[1]*x+M[4]*y+M[7]*z; 
    Z=M[2]*x+M[5]*y+M[8]*z; 
    
    x=X;                              
    y=Y; 
    z=Z;
}   

 
void Vector3d::MultiplyBy(Matrix4x4 &mat)
{
    float *M=mat.M;
    float X,Y,Z; 
                                 
    X=M[0]*x+M[4]*y+M[8]*z+M[12]; 
    Y=M[1]*x+M[5]*y+M[9]*z+M[13]; 
    Z=M[2]*x+M[6]*y+M[10]*z+M[14]; 
    
    x=X;                              
    y=Y; 
    z=Z;
}

//////////////////////////////////////////////////////////////////
//
// Implementations of vector operators that are intended for ease of use
// but will mostly produce slower code - good for non-time critical places
// where you need to write code like: A=5.0f*(B+C*0.6f-D/2.0f)...
//
//////////////////////////////////////////////////////////////////


Vector3d operator+(Vector3d A,Vector3d B)
{
    return Vector3d(A.x+B.x,A.y+B.y,A.z+B.z);   
}


Vector3d operator-(Vector3d A,Vector3d B)
{
    return Vector3d(A.x-B.x,A.y-B.y,A.z-B.z);   
}


Vector3d operator*(Vector3d A,float B)
{
    return Vector3d(A.x*B,A.y*B,A.z*B);  
}


Vector3d operator*(float B,Vector3d A)
{
    return Vector3d(A.x*B,A.y*B,A.z*B);  
}


Vector3d operator/(Vector3d A,float B)
{
    return Vector3d(A.x/B,A.y/B,A.z/B);  
}


//////////////////////////////////////////////////
//
// Matrix3x3 member implementations...
//
///////////////////////////////////////////////////


void Matrix3x3::LoadIdentity()
{
    M[0]=1.0f; M[1]=0.0f; M[2]=0.0f;
    M[3]=0.0f; M[4]=1.0f; M[5]=0.0f;
    M[6]=0.0f; M[7]=0.0f; M[8]=1.0f;
}
    
    
void Matrix3x3::Transpose()
{
    float Temp;
    Temp=M[1]; M[1]=M[3]; M[3]=Temp;
    Temp=M[2]; M[2]=M[6]; M[6]=Temp;
    Temp=M[5]; M[5]=M[7]; M[7]=Temp;
}

    
void Matrix3x3::Invert()
{
    float MD[9];
    
    MD[0]=M[4]*M[8]-M[5]*M[7];
    MD[1]=M[3]*M[8]-M[5]*M[6];
    MD[2]=M[3]*M[7]-M[4]*M[6];
    
    MD[3]=M[1]*M[8]-M[2]*M[7];
    MD[4]=M[0]*M[8]-M[2]*M[6];
    MD[5]=M[0]*M[7]-M[1]*M[6];
    
    MD[6]=M[1]*M[5]-M[2]*M[4];
    MD[7]=M[0]*M[5]-M[2]*M[3];
    MD[8]=M[0]*M[4]-M[1]*M[3];
    
    float DetM=M[0]*M[4]*M[8]+M[1]*M[5]*M[6]+M[2]*M[3]*M[7]
               -M[0]*M[5]*M[7]-M[1]*M[3]*M[8]-M[2]*M[4]*M[6];
    float Mult=1.0f/DetM;
    
    M[0]=MD[0]*Mult;
    M[3]=-MD[1]*Mult;
    M[6]=MD[2]*Mult;
    
    M[1]=-MD[3]*Mult;
    M[4]=MD[4]*Mult;
    M[7]=-MD[5]*Mult;
    
    M[2]=MD[6]*Mult;
    M[5]=-MD[7]*Mult;
    M[8]=MD[8]*Mult;
}

    
void Matrix3x3::MultiplyBy(Matrix3x3 &mat)
{
    float *m1=mat.M;
    float m2[9];
    
    m2[0]=M[0]; m2[1]=M[1]; m2[2]=M[2];
    m2[3]=M[3]; m2[4]=M[4]; m2[5]=M[5];
    m2[6]=M[6]; m2[7]=M[7]; m2[8]=M[8];
    
    M[0]=m1[0]*m2[0]+m1[1]*m2[3]+m1[2]*m2[6];
    M[1]=m1[0]*m2[1]+m1[1]*m2[4]+m1[2]*m2[7];
    M[2]=m1[0]*m2[2]+m1[1]*m2[5]+m1[2]*m2[8];
    M[3]=m1[3]*m2[0]+m1[4]*m2[3]+m1[5]*m2[6];
    M[4]=m1[3]*m2[1]+m1[4]*m2[4]+m1[5]*m2[7];
    M[5]=m1[3]*m2[2]+m1[4]*m2[5]+m1[5]*m2[8];
    M[6]=m1[6]*m2[0]+m1[7]*m2[3]+m1[8]*m2[6];
    M[7]=m1[6]*m2[1]+m1[7]*m2[4]+m1[8]*m2[7];
    M[8]=m1[6]*m2[2]+m1[7]*m2[5]+m1[8]*m2[8];    
}

    
void Matrix3x3::Convert(Quaternion &quat)
{
    float x=quat.x;
    float y=quat.y;
    float z=quat.z;
    float w=quat.w;
    
    float sqw = w*w;    
    float sqx = x*x;    
    float sqy = y*y;    
    float sqz = z*z;
        
    M[0] =  sqx - sqy - sqz + sqw; 
    M[4] = -sqx + sqy - sqz + sqw;    
    M[8] = -sqx - sqy + sqz + sqw; 
           
    float tmp1 = x*y;    
    float tmp2 = z*w;
    M[3] = 2.0 * (tmp1 + tmp2);   
    M[1] = 2.0 * (tmp1 - tmp2);
            
    tmp1 = x*z;
    tmp2 = y*w;    
    M[6] = 2.0 * (tmp1 - tmp2);    
    M[2] = 2.0 * (tmp1 + tmp2);
        
    tmp1 = y*z;    
    tmp2 = x*w;    
    M[7] = 2.0 * (tmp1 + tmp2);    
    M[5] = 2.0 * (tmp1 - tmp2);     
}    


//////////////////////////////////////////////////
//
// Matrix4x4 member implementations...
//
///////////////////////////////////////////////////


void Matrix4x4::LoadIdentity()
{
    M[0]=1.0f; M[1]=0.0f; M[2]=0.0f; M[3]=0.0f;
    M[4]=0.0f; M[5]=1.0f; M[6]=0.0f; M[7]=0.0f;
    M[8]=0.0f; M[9]=0.0f; M[10]=1.0f; M[11]=0.0f;
    M[12]=0.0f; M[13]=0.0f; M[14]=0.0f; M[15]=1.0f;
}

    
void Matrix4x4::Transpose()
{
    float Temp;
    Temp=M[1]; M[1]=M[4]; M[4]=Temp;
    Temp=M[2]; M[2]=M[8]; M[8]=Temp;
    Temp=M[3]; M[3]=M[12]; M[12]=Temp;
    Temp=M[6]; M[6]=M[9]; M[9]=Temp;
    Temp=M[7]; M[7]=M[13]; M[13]=Temp;
    Temp=M[11]; M[11]=M[14]; M[14]=Temp;
}

    
void Matrix4x4::Invert()
{
    float tmp[12];
    float src[16];
    float det;

    for ( int i = 0; i < 4; i++) {
    src[i] = M[i*4];
    src[i + 4] = M[i*4 + 1];
    src[i + 8] = M[i*4 + 2];
    src[i + 12] = M[i*4 + 3];
    }

    tmp[0] = src[10] * src[15];
    tmp[1] = src[11] * src[14];
    tmp[2] = src[9] * src[15];
    tmp[3] = src[11] * src[13];
    tmp[4] = src[9] * src[14];
    tmp[5] = src[10] * src[13];
    tmp[6] = src[8] * src[15];
    tmp[7] = src[11] * src[12];
    tmp[8] = src[8] * src[14];
    tmp[9] = src[10] * src[12];
    tmp[10] = src[8] * src[13];
    tmp[11] = src[9] * src[12];

    M[0] = tmp[0]*src[5] + tmp[3]*src[6] + tmp[4]*src[7];
    M[0] -= tmp[1]*src[5] + tmp[2]*src[6] + tmp[5]*src[7];
    M[1] = tmp[1]*src[4] + tmp[6]*src[6] + tmp[9]*src[7];
    M[1] -= tmp[0]*src[4] + tmp[7]*src[6] + tmp[8]*src[7];
    M[2] = tmp[2]*src[4] + tmp[7]*src[5] + tmp[10]*src[7];
    M[2] -= tmp[3]*src[4] + tmp[6]*src[5] + tmp[11]*src[7];
    M[3] = tmp[5]*src[4] + tmp[8]*src[5] + tmp[11]*src[6];
    M[3] -= tmp[4]*src[4] + tmp[9]*src[5] + tmp[10]*src[6];
    M[4] = tmp[1]*src[1] + tmp[2]*src[2] + tmp[5]*src[3];
    M[4] -= tmp[0]*src[1] + tmp[3]*src[2] + tmp[4]*src[3];
    M[5] = tmp[0]*src[0] + tmp[7]*src[2] + tmp[8]*src[3];
    M[5] -= tmp[1]*src[0] + tmp[6]*src[2] + tmp[9]*src[3];
    M[6] = tmp[3]*src[0] + tmp[6]*src[1] + tmp[11]*src[3];
    M[6] -= tmp[2]*src[0] + tmp[7]*src[1] + tmp[10]*src[3];
    M[7] = tmp[4]*src[0] + tmp[9]*src[1] + tmp[10]*src[2];
    M[7] -= tmp[5]*src[0] + tmp[8]*src[1] + tmp[11]*src[2];

    tmp[0] = src[2]*src[7];
    tmp[1] = src[3]*src[6];
    tmp[2] = src[1]*src[7];
    tmp[3] = src[3]*src[5];
    tmp[4] = src[1]*src[6];
    tmp[5] = src[2]*src[5];
    tmp[6] = src[0]*src[7];
    tmp[7] = src[3]*src[4];
    tmp[8] = src[0]*src[6];
    tmp[9] = src[2]*src[4];
    tmp[10] = src[0]*src[5];
    tmp[11] = src[1]*src[4];

    M[8] = tmp[0]*src[13] + tmp[3]*src[14] + tmp[4]*src[15];
    M[8] -= tmp[1]*src[13] + tmp[2]*src[14] + tmp[5]*src[15];
    M[9] = tmp[1]*src[12] + tmp[6]*src[14] + tmp[9]*src[15];
    M[9] -= tmp[0]*src[12] + tmp[7]*src[14] + tmp[8]*src[15];
    M[10] = tmp[2]*src[12] + tmp[7]*src[13] + tmp[10]*src[15];
    M[10]-= tmp[3]*src[12] + tmp[6]*src[13] + tmp[11]*src[15];
    M[11] = tmp[5]*src[12] + tmp[8]*src[13] + tmp[11]*src[14];
    M[11]-= tmp[4]*src[12] + tmp[9]*src[13] + tmp[10]*src[14];
    M[12] = tmp[2]*src[10] + tmp[5]*src[11] + tmp[1]*src[9];
    M[12]-= tmp[4]*src[11] + tmp[0]*src[9] + tmp[3]*src[10];
    M[13] = tmp[8]*src[11] + tmp[0]*src[8] + tmp[7]*src[10];
    M[13]-= tmp[6]*src[10] + tmp[9]*src[11] + tmp[1]*src[8];
    M[14] = tmp[6]*src[9] + tmp[11]*src[11] + tmp[3]*src[8];
    M[14]-= tmp[10]*src[11] + tmp[2]*src[8] + tmp[7]*src[9];
    M[15] = tmp[10]*src[10] + tmp[4]*src[8] + tmp[9]*src[9];
    M[15]-= tmp[8]*src[9] + tmp[11]*src[10] + tmp[5]*src[8];

    det=src[0]*M[0]+src[1]*M[1]+src[2]*M[2]+src[3]*M[3];

    det = 1/det;
    for ( int j = 0; j < 16; j++)
    M[j] *= det;    
}

    
void Matrix4x4::MultiplyBy(Matrix4x4 &mat)
{
    float *m1=mat.M;
    float m2[16];
    
    m2[0]=M[0]; m2[1]=M[1]; m2[2]=M[2]; m2[3]=M[3];
    m2[4]=M[4]; m2[5]=M[5]; m2[6]=M[6]; m2[7]=M[7];
    m2[8]=M[8]; m2[9]=M[9]; m2[10]=M[10]; m2[11]=M[11];
    m2[12]=M[12]; m2[13]=M[13]; m2[14]=M[14]; m2[15]=M[15];
    
    M[0] = m1[0]*m2[0] + m1[1]*m2[4] + m1[2]*m2[8] + m1[3]*m2[12];
    M[1] = m1[0]*m2[1] + m1[1]*m2[5] + m1[2]*m2[9] + m1[3]*m2[13];
    M[2] = m1[0]*m2[2] + m1[1]*m2[6] + m1[2]*m2[10] + m1[3]*m2[14];
    M[3] = m1[0]*m2[3] + m1[1]*m2[7] + m1[2]*m2[11] + m1[3]*m2[15];
    M[4] = m1[4]*m2[0] + m1[5]*m2[4] + m1[6]*m2[8] + m1[7]*m2[12];
    M[5] = m1[4]*m2[1] + m1[5]*m2[5] + m1[6]*m2[9] + m1[7]*m2[13];
    M[6] = m1[4]*m2[2] + m1[5]*m2[6] + m1[6]*m2[10] + m1[7]*m2[14];
    M[7] = m1[4]*m2[3] + m1[5]*m2[7] + m1[6]*m2[11] + m1[7]*m2[15];
    M[8] = m1[8]*m2[0] + m1[9]*m2[4] + m1[10]*m2[8] + m1[11]*m2[12];
    M[9] = m1[8]*m2[1] + m1[9]*m2[5] + m1[10]*m2[9] + m1[11]*m2[13];
    M[10] = m1[8]*m2[2] + m1[9]*m2[6] + m1[10]*m2[10] + m1[11]*m2[14];
    M[11] = m1[8]*m2[3] + m1[9]*m2[7] + m1[10]*m2[11] + m1[11]*m2[15];
    M[12] = m1[12]*m2[0] + m1[13]*m2[4] + m1[14]*m2[8] + m1[15]*m2[12];
    M[13] = m1[12]*m2[1] + m1[13]*m2[5] + m1[14]*m2[9] + m1[15]*m2[13];
    M[14] = m1[12]*m2[2] + m1[13]*m2[6] + m1[14]*m2[10] + m1[15]*m2[14];
    M[15] = m1[12]*m2[3] + m1[13]*m2[7] + m1[14]*m2[11] + m1[15]*m2[15];
}    

//////////////////////////////////////////////////
//
// Quaternion member implementations...
//
///////////////////////////////////////////////////


float Quaternion::Magnitude()
{
    return sqrtf(x*x+y*y+z*z+w*w);    
}

    
void Quaternion::Normalize()
{
    float mult=1.0f/sqrtf(x*x+y*y+z*z+w*w);
    x*=mult;
    y*=mult;
    z*=mult;
    w*=mult;
}

    
void Quaternion::Invert()
{
    x=-x;
    y=-y;
    z=-z;
}

    
void Quaternion::MultiplyBy(Quaternion &quat)
{
    float Q1x=quat.x;
    float Q1y=quat.y;
    float Q1z=quat.z;
    float Q1w=quat.w;

    float Q2x=x;
    float Q2y=y;
    float Q2z=z;
    float Q2w=w;    
    
    w=Q1w*Q2w-Q1x*Q2x-Q1y*Q2y-Q1z*Q2z;
    x=Q1w*Q2x+Q1x*Q2w+Q1y*Q2z-Q1z*Q2y;
    y=Q1w*Q2y+Q1y*Q2w+Q1z*Q2x-Q1x*Q2z;
    z=Q1w*Q2z+Q1z*Q2w+Q1x*Q2y-Q1y*Q2x;
}

    
void Quaternion::Convert(Matrix3x3 &mat)
{
    float *M=mat.M;
    w = sqrtf(1.0f + M[0] + M[4] + M[8]) * 2.0f;
    x = (M[7] - M[5]) / w ;
    y = (M[2] - M[6]) / w ;
    z = (M[3] - M[1]) / w ;
    w/=4.0f;    
}    


void Quaternion::Slerp(Quaternion &quat,float t)
{
    float Bx=quat.x;
    float By=quat.y;
    float Bz=quat.z;
    float Bw=quat.w;
    float Rez = (x * Bx) + (y * By) + (z * Bz) + (w * Bw);
    
    if(Rez < 0.0f)
    {
        Bx=-Bx;
        By=-By;
        Bz=-Bz;
        Bw=-Bw;
        Rez = -Rez;
    }
    
    float Scale0 = 1.0f - t, Scale1 = t;
    
   	if(1.0f - Rez > 0.1f)
   	{
  		float theta = acosf(Rez);
  		float sinTheta = sinf(theta);
    
  		Scale0 = sinf( ( 1.0f - t ) * theta) / sinTheta;
  		Scale1 = sinf( ( t * theta) ) / sinTheta;
   	}	
    
   	x = (Scale0 * x) + (Scale1 * Bx);
   	y = (Scale0 * y) + (Scale1 * By);
   	z = (Scale0 * z) + (Scale1 * Bz);
   	w = (Scale0 * w) + (Scale1 * Bw);    
}

    
void Quaternion::Slerp(Quaternion &quat1,Quaternion &quat2,float t)
{
    float Ax=quat1.x;
    float Ay=quat1.y;
    float Az=quat1.z;
    float Aw=quat1.w;
    
    float Bx=quat2.x;
    float By=quat2.y;
    float Bz=quat2.z;
    float Bw=quat2.w;
    
    float Rez = (Ax * Bx) + (Ay * By) + (Az * Bz) + (Aw * Bw);
    
    if(Rez < 0.0f)
    {
        Bx=-Bx;
        By=-By;
        Bz=-Bz;
        Bw=-Bw;
        Rez = -Rez;
    }
    
    float Scale0 = 1.0f - t, Scale1 = t;

	if(1.0f - Rez > 0.1f)
	{
		float theta = acosf(Rez);
		float sinTheta = sinf(theta);

		Scale0 = sinf( ( 1.0f - t ) * theta) / sinTheta;
		Scale1 = sinf( ( t * theta) ) / sinTheta;
	}	

	x = (Scale0 * Ax) + (Scale1 * Bx);
	y = (Scale0 * Ay) + (Scale1 * By);
	z = (Scale0 * Az) + (Scale1 * Bz);
	w = (Scale0 * Aw) + (Scale1 * Bw);   
}    
